# Hackish routines for an image manipulator that can do text as a cachable
# image - so it can be composited on things.  And so I can get into the
# surfaces to easily blur them.

# imtext.py is copyright 2014 Spiky Caterpillar and is released under
# the WTFPL.

import wires
import renpy
import pygame
import os
pygame.font.init()

class imText(renpy.display.im.ImageBase):
    chars=(
        u' !"#$%&\'()*+,-./0123456789:;<=>?'+
        u'@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_'+
        u'`abcdefghijklmnopqrstuvwxyz{|}~'+
        u'\u2013\u2014\u2018\u2019\u201c\u201d\u2022\u2030')
    def __init__(self, txt, font, size, color, **properties):
        super(imText, self).__init__(txt, font, color,**properties)
        self.blur = properties.pop('blur',None)
        self.chars = properties.pop('chars',imText.chars)
        self.fallback = properties.pop('fallback',None)
        # Note that vspace handling may not be perfectly accurate in the
        # event that lines vary in height.
        self.vspace = properties.pop('vspace',0)
        self.font = font
        self.txt = txt
        self.size = size
        self.color = color

    def linesurf(self, txt, size=None):
        if txt.endswith('\n'):
            txt = txt[:-1]
        split = []
        a = ''
        ok = True
        fallback_ttf = False
        for c in txt:
            if c in self.chars:
                if ok:
                    a+=c
                else:
                    split.append(a)
                    a = c
                    ok = True
            elif ok:
                split.append(a)
                a = c
                ok = False
                fallback_ttf = True
            else:
                a += c
        if a:
            split.append(a)

        if size is None:
            size = self.size

        ttf = pygame.font.Font(os.path.join(renpy.config.searchpath[0], self.font),size)
        if fallback_ttf:
            fallback_ttf = pygame.font.Font(os.path.join(renpy.config.searchpath[0], self.fallback),size)
        surfs = []
        w = 0
        h = 0
        for chunk in split:
            if chunk[0] in self.chars:
                surf = ttf.render(chunk, 1, self.color)
            else:
                surf = fallback_ttf.render(chunk, 1, self.color)
            h = max(h, surf.get_height())
            w += surf.get_width()
            surfs.append(surf)
        ret = renpy.display.pgrender.surface((w,h),True)
        x = 0
        for chunk in surfs:
            sw, sh = chunk.get_size()
            ret.blit(chunk, (x, (h-sh)>>1))
            x += sw
        return ret

    def load(self, unscaled=False):
        if '\n' in self.txt:
            txts = []
            for line in self.txt.split('\n'):
                txts.append(self.linesurf(line,size=self.size-2))
            widest = 0
            totalh = 0
            for line in txts:
                w,h = line.get_size()
                totalh += h
                widest = max(widest,w)
            surf = renpy.display.pgrender.surface((widest, totalh+((len(txts)-1)*int(self.vspace*h))),True)
            y = 0
            for line in txts:
                w, h = line.get_size()
                surf.blit(line, ((widest-w)>>1, y))
                y += h
                y += int(h * self.vspace)
        else:
            surf = self.linesurf(self.txt)
        if self.blur is not None:
            fuzzed = wires.fuzz(surf, 8, (self.color[0], self.color[1], self.color[2],1))
            fw, fh = fuzzed.get_size()
            tw, th = surf.get_size()
            ret = renpy.display.pgrender.surface(fuzzed.get_size(),True)
            ret.blit(surf,((fw-tw)>>1,(fh-th)>>1))
            ret.blit(fuzzed,(0,0))
        else:
            ret = renpy.display.pgrender.surface(surf.get_size(),True)
            ret.blut(surf,(0,0))
        return ret
