# This is another part of the nopickles system - the scrubber routines.
# Some of them are specific to Date Warp.  Some of them are specific to
# Ren'Py.  Some of them would make sense in any Python code.

# sanitize.py is copyright 2014 Spiky Caterpillar and released under the WTFPL.
# Functions should put debug messages in nopickles.warnings; a
# nopickles.warnings entry is a tuple of the form (severity, data...).
# data can be arbitrary; severity is an int.

# The model I'm using is:
# 0 - minor debug messages.
# 5 - discarded something.
# 6 - accepted something in a manner that may be insecure.

import nopickles
import renpy
import animation
import namedtransform

def const(constval):
    def const_ret(s):
        if type(s) is not type(constval) or s!=constval:
            nopickles.warnings.append((5,'Expected constant', constval,'got a',type(s),s))
        return constval
    return const_ret

def discard(s):
    return None

def expect_Character_var_name_or_None(s):
    if s is None:
        return None
    if isinstance(getattr(renpy.store,s), renpy.character.ADVCharacter):
        # Also accept child classes of ADVCharacter, so NVLCharacter and
        # other weird custom variants will work.
        return s
    nopickles.warnings.append((5,"Expected an ADVCharacter, discarded",type(s),s))
    return None

def scrub_last_say_who(s):
    if s is None:
        return None
    if type(s) is str and s[0]=='"' and s[-1]=='"':
        # Sometimes who is just a barestring.
        return s
    if isinstance(getattr(renpy.store,s), renpy.character.ADVCharacter):
        # Also accept child classes of ADVCharacter, so NVLCharacter and
        # other weird custom variants will work.
        return s
    nopickles.warnings.append((5,"Expected an ADVCharacter, discarded",type(s),s))
    return None

def expect_False(s):
    if s is not False:
        nopickles.warnings.append((5,'Expected False, got a',type(s)))
    return False

def expect_None(s):
    if s is not None:
        nopickles.warnings.append((5, 'Expected None, got a',type(s)))
    return None

def expect_bool(s):
    if type(s) is not bool:
        nopickles.warnings.append((5, 'Expected a bool, got',type(s),s))
        return False
    return s

def expect_bool_or_del(s):
    if type(s) is bool:
        return s
    if type(s) is not renpy.python.StoreDeleted:
        nopickles.warnings.append((5,'expect_bool_or_del discarded',type(s),s))
    return renpy.python.deleted

def expect_bool_or_None(s):
    if type(s) is bool:
        return s
    if s is not None:
        nopickles.warnings.append((5,'expect_bool_or_None discarded',type(s),s))
    return None

def expect_char_or_del(s):
    if type(s) is str and len(s)==1:
        return s
    if type(s) is not renpy.python.StoreDeleted:
        nopickles.warnings.append((5,'expect_char_or_del discarded',type(s),s))
    return renpy.python.deleted

def expect_del(s):
    if type(s) is not renpy.python.StoreDeleted:
        nopickles.warnings.append((5, 'Expected a deleted variable, got ', type(s),s))
    return renpy.python.deleted

def expect_del_or_emptydict(s):
    if type(s) is dict and not s:
        return {}
    if type(s) is not renpy.python.StoreDeleted:
        nopickles.warnings.append((5,'expect_del_or_emptydict discarded',type(s),s))
    return renpy.python.deleted

def expect_del_or_emptytuple(s):
    if type(s) is tuple and not s:
        return ()
    if type(s) is not renpy.python.StoreDeleted:
        nopickles.warnings.append((5, 'expect_del_or_emptytuple discarded', type(s),s))
    return renpy.python.deleted

def expect_del_or_int(s):
    if type(s) is int:
        return s
    if type(s) is not renpy.python.StoreDeleted:
        nopickles.warnings.append((5, 'expect_del_or_int discarded', type(s),s))
    return renpy.python.deleted

def expect_del_or_int_or_None(s):
    if type(s) is int:
        return s
    if s is None:
        return s
    if type(s) is not renpy.python.StoreDeleted:
        nopickles.warnings.append((5, 'expect_del_or_int discarded', type(s),s))
    return renpy.python.deleted

def expect_del_or_int2(s):
    if type(s) is tuple and len(s)==2 and type(s[0]) is int and type(s[1]) is int:
        return s
    elif type(s) is not renpy.python.StoreDeleted:
        nopickles.warnings.append((5,"Expected (int, int) or deleted, got", type(s)))
    return renpy.python.deleted

def expect_del_or_str(s):
    if type(s) is str:
        return s
    if type(s) is not renpy.python.StoreDeleted:
        nopickles.warnings.append((5,'expect_del_or_str discarded',type(s),s))
    return renpy.python.deleted

OK_SIMPLE_TEXT_TAGS = set(('/a','b','/b','fast','i','/i','nw','/size'))
# XXX This is NOT an exhaustive list of tags.  You should search your script
# for all tags used and scrub them accordingly (keeping in mind that some
# of them may actually have security issues!)

# URLs which may be used with the {a=} tag.  This needs to be a datatype that
# the 'in' operator works on, hence the trailing comma.
OK_URLS = 'http://www.hanakogames.com/datewarp.shtml',

def expect_displayable_text(s):
    if type(s) is not str and type(s) is not unicode:
        nopickles.warnings.append((5, 'expected a string or unicode, got', type(s)))
        return '-error.31-'
    ret = ''
    i = 0
    while i<len(s):
        if s[i]=='{':
            split_s = s[i+1:].split('}',1)
            if len(split_s)!=2:
                raise Exception("Incomplete tag.")
            if split_s[0] in OK_SIMPLE_TEXT_TAGS:
                ret += '{'+split_s[0]+'}'
                i = 0
                s = split_s[1]
            elif '=' in split_s[0]:
                tag, var = split_s[0].split('=',1)
                if tag=='a' and var in OK_URLS:
                    ret += '{a='+var+'}'
                    s = split_s[1]
                    i = 0
                elif tag=='size':
                    ret += '{size='+str(int(var))+'}'
                    s = split_s[1]
                    i = 0
                elif tag=='w':
                    ret += '{w='+str(float(var))+'}'
                    s = split_s[1]
                    i = 0
                else:
                    nopickles.warnings.append((5,'Unhandled complex tag',s,split_s))
                    print s
                    raise Exception('Unhandled complex tag in saved data.')
            else:
                # XXX at release time, warn and scrub instead of crashing.
                print "Debug: 26:",s
                raise Exception("Unhandled text tag in saved data.")
        else:
            ret += s[i]
            i += 1
    return ret

def expect_displayable_text_or_None(s):
    if s is None:
        return None
    return expect_displayable_text(s)

def expect_displayable_text_or_del_or_None(s):
    if type(s) is renpy.python.StoreDeleted:
        return renpy.python.deleted
    return expect_displayable_text_or_None(s)

def expect_float_or_int(s):
    if type(s) is float or type(s) is int:
        return s
    nopickles.warnings.append((5,'expect_float_or_int discarded',s))
    return 0

def expect_float_or_None(s):
    if type(s) is float:
        return s
    if s is not None:
        nopickles.warnings.append((5,'expect_float_or_None discarded',type(s),s))
    return None

def expect_int(s):
    if type(s) is not int:
        nopickles.warnings.append((5,'Expected an int, discarded',type(s),s))
        return 0
    return s

def expect_int_or_None(s):
    if type(s) is int:
        return s
    if s is not None:
        nopickles.warnings.append((5,'expect_int_or_None discarded',type(s),s))
    return None

def expect_str(s):
    if type(s) is not str:
        if nopickles.INSECURE:
            nopickles.warnings.append((6,'Expected str, accepted',type(s),s))
            return s
        nopickles.warnings.append((5,'Expected str, got a',type(s)))
        return ''
    return s

def expect_str_or_None(s):
    if s is None:
        return None
    if type(s) is not str:
        nopickles.warnings.append((5,'Expected a str or None, discarded',type(s),s))
        return None
    return s

# new_* sanitizers will silently replace the save's contents.
# This is largely to purge the rather pointless prediction of the main menu
# from inside the game.
def new_empty_revertable_list(s):
    return renpy.python.RevertableList()

# scrub_* do more detailed processing on things and are much more likely
# to be game-specific.
def scrub__game_menu_screen(s):
    # For _game_menu_screen and _old_game_menu_screen, which are used
    # internally by Ren'Py.
    if type(s) is str:
        if s in ('preferences_screen','save'):
            return s
    elif s is None:
        return None
    elif type(s) is renpy.python.StoreDeleted:
        return renpy.python.deleted
    if nopickles.INSECURE:
        nopickles.warnings.append((6,'Accepted _game_menu_screen',type(s),s))
    else:
        raise Exception("Unexpected _game_menu_screen.")

def scrub_dimension_anim_base_profiles(s):
    if type(s) is renpy.python.StoreDeleted:
        return renpy.python.deleted
    # Very DateWarp-specific.  May need to track changes to effects.rpy's
    # DimensionAnimDisplayable() code.
    ret = renpy.python.RevertableList()
    for ent in s:
        # Each entry in dimension_anim_base_profiles may be:
        # None,
        # (True, alpha),
        # (start_t, dur, start_a, end_a, ease)
        if ent is None:
            ret.append(None)
        elif ent[0] is True:
            ret.append((True,int(ent[1])))
        elif isinstance(ent[0], animation.AnimController):
            ret.append((ent[0], expect_float_or_None(ent[1])))
        else:
            print ret,ent
            ret.append((expect_float_or_None(ent[0]),expect_float_or_None(ent[1]), expect_int_or_None(ent[2]), int(ent[3]), bool(ent[4])))
    return ret

# This is DateWarp specific.
NVL_PROPS = dict(
    say_vbox_properties=dict(box_layout='horizontal'),
    window_args=dict(style='nvl_entry'),
    who_args=dict(style='nvl_label', substitute=False),
    what_args=dict(style='nvl_dialogue', substitute=False, slow_abortable=True))
def scrub_nvl_list(s):
    if s is None:
        return s
    scrubbed = []
    for ent in s:
        if len(ent)!=3:
            nopickles.warnings.append((5,'nvl_list entry not a 3-tuple.'))
        if ent[0] is not None:
            nopickles.warnings.append((5,'nvl_list entry[0] not None.'))
        if ent[2]!=NVL_PROPS:
            nopickles.warnings.append((5,'nvl_list entry[2] had unexpected properties',ent[2]))
        scrubbed.append((None, expect_displayable_text(ent[1]), NVL_PROPS))
    return renpy.python.RevertableList(scrubbed)

def scrub_persistent_seen_audio(s):
    if type(s) is not dict:
        nopickles.warnings.append((5,'discarded invalid _seen_audio',type(s)))
        return {}
    ret = {}
    for k,v in s.iteritems():
        if type(k) is str and v is True:
            ret[k] = True
        else:
            nopickles.warnings.append((5,'discarded _seen_audio member',k,v))
    return ret

OK_JOYMAP_KEYS = ('joy_left','joy_right','joy_up','joy_down',
    'joy_dismiss',
    'joy_rollback', 'joy_hide', 'joy_holdskip', 'joy_menu',
    'joy_toggleskip')

def ok_joyitem_number(n):
    # Joyitems are assumed to be (device).(widgetnum), where both device
    # and widgetnums are ints.  Bounds checking here should be unnecessary -
    # OS-supplied joystick events are read and converted to strings matching
    # the joystick format.
    n = n.split('.')
    if len(n)==2 and n[0]==str(int(n[0])) and n[1]==str(int(n[1])):
        return True
    return False

def ok_joymap_value(s):
    if type(s) is not str:
        return False
    s = s.split(' ')
    if s[0]=='Axis' and len(s)==3 and ok_joyitem_number(s[1]) and s[2] in ('Negative','Positive'):
        return True
    elif s[0] == 'Button' and ok_joyitem_number(s[1]) and len(s)==2:
        return True
    return False

def scrub_preference_joymap(s):
    if type(s) is dict:
        ret = {}
        for k,v in s.iteritems():
            if k in OK_JOYMAP_KEYS and ok_joymap_value(v):
                ret[k] = v
            else:
                nopickles.warnings.append((5,'discarded unsupported joymap entry',k,v))
        return ret
    nopickles.warnings.append((5,'discarded unsupported joymap',type(s),s))
    return {}

def scrub_preference_mute(s):
    if type(s) is not dict:
        nopickles.warnings.append((5,'discarded unsupported mute',type(s),s))
        return {}
    ret = {}
    for k,v in s.iteritems():
        if type(k) is str and type(v) is bool:
            ret[k] = v
        else:
            nopickles.warnings.append((5,'discarded unsupported mute member',k,v))
    return ret

def scrub_preference_physical_size(s):
    if type(s) is tuple and len(s)==2 and type(s[0]) is int and type(s[1]) is int:
        return s
    nopickles.warnings.append((5,'discarded invalid physical size',s))
    return None

OK_RENDERERS = ('auto','angle','gl','sw')
def scrub_preference_renderer(s):
    if s in OK_RENDERERS:
        return str(s)
    nopickles.warnings.append((5,'discarded unsupported renderer',s))
    return 'auto'

def scrub_preference_volumes(s):
    if type(s) is not dict:
        nopickles.warnings.append((5,'discarded unsupported volumes',s))
        return {}
    ret = {}
    for k,v in s.iteritems():
        if type(k) is str and (type(v) is float or type(v) is int):
            ret[k] = v
        else:
            nopickles.warnings.append((5,'discarded unsupported volume',k,v))
    return ret

OK_FILENAME_CHARS = '-.0123456789_abcdefghijklmnopqrstuvwxyz'
def scrub_relative_path(s):
    # This assumes that the path is already a string or unicode, and that
    # we're not really worried about anything but invalid characters and
    # attempts to crawl out of directories.
    path = s.split('/')
    for component in path:
        if component in ('','.','..'):
            raise Exception('Dodgy entry "'+component+'" in relative path.')
        for c in component:
            if c not in OK_FILENAME_CHARS:
                raise Exception("Disallowed character in component: "+ord(c))
    return str(s)

def scrub_spritegroup(s):
    if type(s) is renpy.python.StoreDeleted:
        return renpy.python.deleted
    if s is None:
        return None
    ret = renpy.python.RevertableList()
    try:
        for ent in s:
            if ent[0] in ('alben', 'bradley','linds','nathaniel','rafael'):
                if type(ent[1]) is namedtransform.NamedTransform:
                    if len(ent)>2 and type(ent[2]) is not str:
                        nopickles.warnings.append((5, "scrub_spritegroup truncated",ent))
                        ent = ent[:2]
                    ret.append(renpy.python.RevertableList(ent[:3]))
                elif type(ent[1]) is str and len(ent)==2:
                    ret.append(renpy.python.RevertableList(ent))
                else:
                    nopickles.warnings.append((5, "scrub_spritegroup discarded",ent))
            else:
                nopickles.warnings.append((5, "scrub_spritegroup discarded",ent))
        return ret
    except Exception,e:
        nopickles.warnings.append((5,"scrub_spritegroup got exception",e))
    if nopickles.INSECURE:
        nopickles.warnings.append((6,"Accepted spritegroup",type(s),s))
        return s
    else:
        raise Exception("scrub_spritegroup unfinished XXX")
